/* Copyright 2007 Cory McWilliams <cory@unprompted.com>
 * Based on convmvfs 2006 by ZC Miao <hellwolf.misty@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 */

#define FUSE_USE_VERSION 25
#define VERSION "0.1"
#include <fuse.h>
#include <fuse_opt.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cstddef>
#include <cassert>
#include <string>

using namespace std;

/*
 * global vars
 */
static const char* CIMIRRORFS_DEFAULT_SRCDIR = "/"; /* root dir */

struct cimirrorfs {
	const char *cwd;
	const char *srcdir;
};
static struct cimirrorfs cimirrorfs;

static uid_t euid;
gid_t egid;

static void init_gvars(){
	cimirrorfs.cwd = get_current_dir_name();
	cimirrorfs.srcdir = CIMIRRORFS_DEFAULT_SRCDIR;

	euid = geteuid();
	egid = getegid();
}

static struct fuse_operations cimirrorfs_oper;

/*
 * options and usage
 */
enum {
	KEY_HELP,
	KEY_VERSION,
};

#define CIMIRRORFS_OPT(t, p, v) { t, offsetof(struct cimirrorfs, p), v }

static struct fuse_opt cimirrorfs_opts[] = {
	CIMIRRORFS_OPT("srcdir=%s", srcdir, 0),

	FUSE_OPT_KEY("-V",				KEY_VERSION),
	FUSE_OPT_KEY("--version", KEY_VERSION),
	FUSE_OPT_KEY("-h",				KEY_HELP),
	FUSE_OPT_KEY("--help",		KEY_HELP),
	{NULL,0,0}
};

static void usage(){
	printf(
				 "CIMIRRORFS options:\n"
				 "		-o srcdir=PATH				 which directory to convert\n"
				 );
}

static int cimirrorfs_opt_proc(void *data, const char *arg,int key,
													struct fuse_args *outargs)
{
	(void)arg;
	(void)data;
	
	switch (key) {
	case FUSE_OPT_KEY_OPT:
	case FUSE_OPT_KEY_NONOPT:
		return 1;
	case KEY_VERSION:
		fprintf(stderr, "cimirror\t"VERSION"\n"
						"Copyright 2007 Cory McWilliams <cory@unprompted.com>\n\n"
						"This program is free software; you can redistribute it and/or modify\n"
						"it under the terms of the GNU General Public License version 3 as\n"
						"published by the Free Software Foundation.\n");
		exit(0);
	case KEY_HELP:
		fuse_opt_add_arg(outargs, "-h");
		fuse_main(outargs->argc, outargs->argv, &cimirrorfs_oper);
		usage();
		exit(0);
	default:
		fprintf(stderr, "unknown option\n");
		exit(1);
	}
}



/*
 * util funs
 */
#define OUTINBUFLEN 255

std::string out2in(const char* s) {
	std::string full = cimirrorfs.srcdir + std::string(s);
	size_t pos = std::strlen(cimirrorfs.srcdir);
	while ((pos = full.find('/', pos)) != std::string::npos) {
		size_t next = full.find('/', pos + 1);
		if (next == std::string::npos) {
			next = full.size();
		}
		std::string here = full.substr(0, pos + 1);
		std::string chunk = full.substr(pos + 1, next - pos - 1);
		DIR *dir = opendir(here.c_str());
		if (dir == NULL) {
			return full;
		}
		struct dirent *ent;
		while ((ent = readdir(dir)) != NULL) {
			if (!strcasecmp(ent->d_name, chunk.c_str())) {
				std::memcpy(&*full.begin() + pos + 1, ent->d_name, chunk.size());
				break;
			}
		}
		if (closedir(dir)) {
			perror("closedir failed");
			return full;
		}
		pos++;

	}
	printf("%s => %s\n", s, full.c_str());
	return full;
}


#define PERM_WALK_CHECK_READ	 01
#define PERM_WALK_CHECK_WRITE	02
#define PERM_WALK_CHECK_EXEC	 04
static int permission_walk(const char *path, uid_t uid, gid_t gid,
													 int perm_chk, int readlink = 0){
	int rt;
	//I'm root~~
	if(uid == 0){
		return 0;
	}
	size_t l = strlen(path) + 1;
	char *p = (char*)malloc(l);
	if(p == NULL){
		return -ENOMEM;
	}
	strncpy(p, path, l);

	char *s = p;
	if(*s == '\0'){
		//Empty pathname, see PATH_RESOLUTION(2)
		rt = -ENOENT;
		goto __free_quit;
	}
	while(*s++){
		struct stat stbuf;
		int chk;
		if(*s == '\0'){
			//final entry
			if(readlink?lstat(p, &stbuf):stat(p, &stbuf)){
				rt = -errno;
				goto __free_quit;
			}
			chk = perm_chk;
		}else if(*s == '/'){
			//non-final component
			*s = '\0';
			if(stat(p, &stbuf)){
				rt =	-errno;
				goto __free_quit;
			}
			if(!(stbuf.st_mode & S_IFDIR)){
				rt = -ENOTDIR;
				goto __free_quit;
			}
			*s = '/';
			chk = PERM_WALK_CHECK_EXEC;
		}else{
			continue;
		}
		int mr,mw,mx;
		if(stbuf.st_uid == uid){
			mr = S_IRUSR;
			mw = S_IWUSR;
			mx = S_IXUSR;
		}else if(stbuf.st_gid == gid){
			mr = S_IRGRP;
			mw = S_IWGRP;
			mx = S_IXGRP;
		}else{
			mr = S_IROTH;
			mw = S_IWOTH;
			mx = S_IXOTH;
		}
		if(chk & PERM_WALK_CHECK_READ){
			if(!(stbuf.st_mode & mr)){
				rt = -EACCES;
				goto __free_quit;
			}
		}
		if(chk & PERM_WALK_CHECK_WRITE){
			if(!(stbuf.st_mode & mw)){
				rt = -EACCES;
				goto __free_quit;
			}
		}
		if(chk & PERM_WALK_CHECK_EXEC){
			if(!(stbuf.st_mode & mx)){
				rt = -EACCES;
				goto __free_quit;
			}
		}
	}

	rt = 0;
__free_quit:
	free(p);
	return rt;
}

static int permission_walk_parent(const char *path, uid_t uid, gid_t gid,
																	int perm_chk){
	int l = strlen(path);
	while(--l)
		if(path[l] == '/')
			break;
	return permission_walk(string(path, l).c_str(), uid, gid,
												 perm_chk);
}


/*
 * opers
 */
static void *cimirrorfs_init(void){
	if(chdir(cimirrorfs.cwd)){
		perror("fuse init,chdir failed");
		exit(errno);
	}
	return NULL;
}

static int cimirrorfs_open(const char *opath, struct fuse_file_info *fi){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st;
	if(fi->flags & O_WRONLY){
		st = permission_walk(ipath.c_str(), cont->uid, cont->gid,
													 PERM_WALK_CHECK_WRITE);
	}else if(fi->flags & O_RDWR){
		st = permission_walk(ipath.c_str(), cont->uid, cont->gid,
												 PERM_WALK_CHECK_READ|PERM_WALK_CHECK_WRITE);
	}else{
		st = permission_walk(ipath.c_str(), cont->uid, cont->gid,
												 PERM_WALK_CHECK_READ);
	}
	if(st)
		return st;

	int fd;
	if( ( fd = open(ipath.c_str(),fi->flags)) == -1 ){
		return -errno;
	}
	fi->fh = fd;

	return 0;
}

static int cimirrorfs_read(const char *opath, char *buf,
												 size_t size, off_t offset,
												 struct fuse_file_info *fi){
	(void)opath;

	lseek(fi->fh, offset, SEEK_SET);
	return read(fi->fh, buf, size);
}

static int cimirrorfs_write(const char *opath, const char *buf, size_t size, off_t off, struct fuse_file_info *fi){
	(void)opath;

	lseek(fi->fh, off, SEEK_SET);
	return write(fi->fh, buf, size);
}

static int cimirrorfs_release(const char *opath, struct fuse_file_info *fi){
	(void)opath;

	if(close(fi->fh))
		return -errno;
	return 0;
}

static int cimirrorfs_getattr(const char *opath, struct stat *stbuf){
	string ipath = out2in(opath);

	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk_parent(ipath.c_str(), cont->uid, cont->gid,
																	PERM_WALK_CHECK_EXEC);
	if(st)
		return st;

	if(lstat(ipath.c_str(), stbuf)){
		return -errno;
	}
	return 0;
}

static int cimirrorfs_opendir(const char *opath, struct fuse_file_info *fi){
	(void)fi;
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk(ipath.c_str(), cont->uid, cont->gid,
													 PERM_WALK_CHECK_READ);
	if(st)
		return st;

	return 0;
}

static int cimirrorfs_readdir(const char *opath, void *buf, fuse_fill_dir_t filler,
												 off_t offset, struct fuse_file_info *fi){
	(void)offset;
	(void)fi;
	string ipath = out2in(opath);

	DIR * dir;
	if( (dir = opendir(ipath.c_str())) == NULL ){
		return -errno;
	}
	struct dirent *pdirent;
	pdirent = readdir(dir);
	while ( pdirent != NULL ) {
		filler(buf, pdirent->d_name,
					 NULL, 0);
		pdirent = readdir( dir );
	}
	closedir(dir);
	return 0;
}

static int cimirrorfs_mknod (const char *opath, mode_t mode, dev_t dev){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk_parent(ipath.c_str(), cont->uid, cont->gid,
																	PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;

	int rt = mknod(ipath.c_str(), mode, dev);
	if(rt)return -errno;
	if(euid == 0){
		chown(ipath.c_str(), cont->uid, cont->gid);
	}
	return 0;
}

static int cimirrorfs_mkdir (const char *opath, mode_t mode){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk_parent(ipath.c_str(), cont->uid, cont->gid,
																	PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;

	int rt = mkdir(ipath.c_str(), mode);
	if(rt)return -errno;
	if(euid == 0){
		chown(ipath.c_str(), cont->uid, cont->gid);
	}
	return 0;
}

static int cimirrorfs_readlink(const char *opath,
														 char *path, size_t path_len){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk(ipath.c_str(), cont->uid, cont->gid,
													 PERM_WALK_CHECK_READ, 1);
	if(st)
		return st;

	st = readlink(ipath.c_str(), path, path_len-1);
	if(st == -1)
		return -errno;
	path[st] = '\0';
	ipath = path;
	strncpy(path, ipath.c_str(), min(path_len,ipath.size()+1));

	return 0;
}

static int cimirrorfs_unlink(const char *opath){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk_parent(ipath.c_str(), cont->uid, cont->gid,
																	PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;

	if(unlink(ipath.c_str()))
		return -errno;
	return 0;
}

static int cimirrorfs_rmdir(const char *opath){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk_parent(ipath.c_str(), cont->uid, cont->gid,
																	PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;

	if(rmdir(ipath.c_str()))
		return -errno;
	return 0;
}

static int cimirrorfs_symlink(const char *oldpath, const char *newpath){
	string inewpath = out2in(newpath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk_parent(inewpath.c_str(), cont->uid, cont->gid,
																	PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;

	if(symlink(out2in(oldpath).c_str(), inewpath.c_str()))
		return -errno;
	return 0;
}

static int cimirrorfs_rename(const char *oldpath, const char *newpath){
	string inewpath = out2in(newpath);
	string ioldpath = out2in(oldpath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk_parent(inewpath.c_str(), cont->uid, cont->gid,
																	PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;
	st = permission_walk_parent(ioldpath.c_str(), cont->uid, cont->gid,
															PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;
	struct stat stbuf;
	if(stat(ioldpath.c_str(), &stbuf)){
		return -errno;
	}
	if(stbuf.st_mode & S_IFDIR){
		/* see rename(2), need write permission for renamed directory, because
		 * needed to update the .. entry
		 */
		st = permission_walk(ioldpath.c_str(), cont->uid, cont->gid,
												 PERM_WALK_CHECK_WRITE);
		if(st)
			return st;
	}

	if(rename(ioldpath.c_str(), inewpath.c_str()))
		return -errno;
	return 0;
}

static int cimirrorfs_link(const char *oldpath, const char *newpath){
	string inewpath = out2in(newpath);
	string ioldpath = out2in(oldpath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk_parent(inewpath.c_str(), cont->uid, cont->gid,
																	PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;
	st = permission_walk_parent(ioldpath.c_str(), cont->uid, cont->gid,
															PERM_WALK_CHECK_WRITE|PERM_WALK_CHECK_EXEC);
	if(st)
		return st;

	if(link(ioldpath.c_str(), inewpath.c_str()))
		return -errno;
	return 0;
}

static int cimirrorfs_chmod(const char *opath, mode_t mode){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	struct stat stbuf;
	if(stat(ipath.c_str(), &stbuf)){
		return -errno;
	}
	if(cont->uid != stbuf.st_uid)
		return -EPERM;
	if(chmod(ipath.c_str(),mode))
		return -errno;
	return 0;
}

static int cimirrorfs_truncate(const char *opath, off_t length){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk(ipath.c_str(), cont->uid, cont->gid,
													 PERM_WALK_CHECK_WRITE);
	if(st)
		return st;

	if(truncate(ipath.c_str(), length))
		return -errno;
	return 0;
}

static int cimirrorfs_utime(const char *opath, struct utimbuf *buf){
	string ipath = out2in(opath);

	/* permission check*/
	struct fuse_context *cont = fuse_get_context();
	struct stat stbuf;
	if(stat(ipath.c_str(), &stbuf)){
		return -errno;
	}

	if(buf == NULL && permission_walk(ipath.c_str(), cont->uid, cont->gid,
																		PERM_WALK_CHECK_WRITE)){
		return -errno;
	}
	if(buf != NULL && cont->uid != stbuf.st_uid)
		return -EPERM;

	if(utime(ipath.c_str(), buf))
		return -errno;
	return 0;
}

static int cimirrorfs_access(const char *opath, int mode){
	string ipath = out2in(opath);

	if(mode & F_OK){
		struct stat stbuf;
		if(stat(ipath.c_str(),&stbuf)){
			return -errno;
		}
	}
	struct fuse_context *cont = fuse_get_context();
	return permission_walk(ipath.c_str(), cont->uid, cont->gid,
												 (mode & R_OK)?PERM_WALK_CHECK_READ:0 |
												 (mode & W_OK)?PERM_WALK_CHECK_WRITE:0 |
												 (mode & X_OK)?PERM_WALK_CHECK_EXEC:0
												 );
}

static int cimirrorfs_chown(const char *opath, uid_t uid_2set, gid_t gid_2set){
	string ipath = out2in(opath);

	struct fuse_context *cont = fuse_get_context();
	if(cont->uid == 0){
		if(chown(ipath.c_str(), uid_2set, gid_2set)){
			return -errno;
		}else{
			return 0;
		}
	}else{
		return -EPERM;
	}
}

static int cimirrorfs_statfs(const char *opath, struct statvfs *buf){
	string ipath = out2in(opath);

	struct fuse_context *cont = fuse_get_context();
	int st = permission_walk(ipath.c_str(), cont->uid, cont->gid,0);
	if(st)
		return st;

	if(statvfs(ipath.c_str(), buf)){
		return -errno;
	}

	return 0;
}

static void cimirrorfs_oper_init(){
	memset(&cimirrorfs_oper, 0, sizeof(cimirrorfs_oper));
	cimirrorfs_oper.getattr = cimirrorfs_getattr;
	cimirrorfs_oper.opendir = cimirrorfs_opendir;
	cimirrorfs_oper.readdir = cimirrorfs_readdir;
	cimirrorfs_oper.readlink = cimirrorfs_readlink;
	cimirrorfs_oper.mknod = cimirrorfs_mknod;
	cimirrorfs_oper.mkdir = cimirrorfs_mkdir;
	cimirrorfs_oper.unlink = cimirrorfs_unlink;
	cimirrorfs_oper.rmdir = cimirrorfs_rmdir;
	cimirrorfs_oper.symlink = cimirrorfs_symlink;
	cimirrorfs_oper.rename = cimirrorfs_rename;
	cimirrorfs_oper.link = cimirrorfs_link;
	cimirrorfs_oper.chmod = cimirrorfs_chmod;
	cimirrorfs_oper.chown = cimirrorfs_chown;
	cimirrorfs_oper.truncate = cimirrorfs_truncate;
	cimirrorfs_oper.utime = cimirrorfs_utime;
	cimirrorfs_oper.open = cimirrorfs_open;
	cimirrorfs_oper.read = cimirrorfs_read;
	cimirrorfs_oper.write = cimirrorfs_write;
	cimirrorfs_oper.release = cimirrorfs_release;
	cimirrorfs_oper.access = cimirrorfs_access;
	cimirrorfs_oper.statfs = cimirrorfs_statfs;

	cimirrorfs_oper.init = cimirrorfs_init;
}


/*
 * life is here
 */
int main(int argc, char *argv[])
{
	int res;
	string srcdir;

	init_gvars();

	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);	
	if (fuse_opt_parse(&args, &cimirrorfs, cimirrorfs_opts, cimirrorfs_opt_proc) == -1)
		exit(1);
	if(strlen(cimirrorfs.srcdir)){
		srcdir = cimirrorfs.srcdir;
		if(srcdir[srcdir.size()-1] == '/'){
			srcdir.erase(srcdir.size()-1,1);
		}
		size_t p;
		while((p = srcdir.find("//")) != string::npos){
			srcdir.erase(p,1);
		}
		cimirrorfs.srcdir = srcdir.c_str();
	}

	cimirrorfs_oper_init();

	fprintf(stderr,
					"srcdir=%s\n",
					cimirrorfs.srcdir);

	res = fuse_main(args.argc, args.argv, &cimirrorfs_oper);

	return res;
}
